#!/usr/bin/env python
# -*- coding: utf-8 -*-

# WRCUI (c) 2008 Michel Blockelet <blockelet@crans.org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.



"""WRCUI -- WebRadio de Cachan - User Interface

WRCUI est un client pour le système de WebRadio du campus de Cachan."""

__author__ = "Michel Blockelet"
__email__ = "blockelet@crans.org"

config_version = '0.1 beta 6'



### Importations
# Bibliothèques courantes (ou pas)
import os, sys, string, time
import pickle
import ConfigParser

# Interface utilisateur
import pygtk
pygtk.require('2.0')
import gtk, gobject

# Communications
import irclib
import socket

# Pour les Threads (enfin, pas sous Windows)
import threading

# Pour la correspondance dans l'EPG
import difflib

# Pour le parsing du fichier de Rhythmbox
import xml.dom.minidom
import urllib

# Pour démarrer le Player
import subprocess


### Options de configuration
config_configfile = os.path.expanduser("~/.wrcui")

# Serveur WRCServ
config_data_server = "webradio.crans.org"
config_data_port = 15000
config_data_timeout = 3
config_data_reconnect = 2

# Serveur IRC
config_irc_server = "irc.crans.org"
config_irc_port = 6667
config_irc_chan = "#wrc"
config_irc_reconnect = 10

# Paramètres de debug
debug_stdout = True
if sys.platform == "win32":
    debug_fileobject = open(os.path.expanduser("~/.wrcui_log.txt"), 'a')
else:
    debug_fileobject = open("/tmp/wrcui.log", 'a')


### Variables globales
debug_lock = threading.Lock()



def debug(text="", name=""):
    """Afficher des informations de debug, ou, si text = \"\", les détails
    d'une exception rattrapée."""
    debugout = time.strftime("%Y/%m/%d %H:%M:%S ")
    if text != "":
        debugout += (name or "Debug") + " :: " + text
    else:
        exctype, value = sys.exc_info()[:2]
        debugout += ((name or "Unknown") + " -- Exception " + str(exctype)
            + " : " + str(value))
    debug_lock.acquire()
    if debug_stdout:
        print debugout
    if debug_fileobject != None:
        try:
            debug_fileobject.write(debugout + "\n")
        except:
            pass
    debug_lock.release()


class Config:
    """Classe de configuration (pour simplifier encore plus ConfigParser !)."""

    def get(self, section, option, default=""):
        """Récupérer une valeur."""
        if self.configparser.has_option(section, option):
            return self.configparser.get(section, option)
        else:
            self.set(section, option, default)
            return default
        return self.configparser.get(section, option)

    def getint(self, section, option, default=0):
        """Récupérer une valeur représentant un entier."""
        if self.configparser.has_option(section, option):
            return self.configparser.getint(section, option)
        else:
            self.set(section, option, default)
            return default

    def getfloat(self, section, option, default=0.0):
        """Récupérer une valeur représentant un flottant."""
        if self.configparser.has_option(section, option):
            return self.configparser.getfloat(section, option)
        else:
            self.set(section, option, default)
            return default

    def getboolean(self, section, option, default=False):
        """Récupérer une valeur représentant un booléen."""
        if self.configparser.has_option(section, option):
            return self.configparser.getboolean(section, option)
        else:
            self.set(section, option, default)
            return default

    def set(self, section, option, value):
        """Affecter une valeur et écrire le fichier de configuration."""
        if not self.configparser.has_section(section):
            self.configparser.add_section(section)
        self.configparser.set(section, option, str(value))
        # Sur la plupart des OS, ce sera seulement inscrit dans le cache disque
        self.configparser.write(open(self.filename, 'w'))

    def __init__(self, filename):
        """Initialise la classe avec un ConfigParser."""
        self.configparser = ConfigParser.ConfigParser()
        self.configparser.read(filename)
        self.filename = filename


class ServerComm:
    """Classe d'abstraction de la communication client/serveur."""

    ### Fonctions utilitaires
    def debug(self, text=""):
        """Affiche des informations de debug."""
        debug(text, "ServerComm")

    ### Fonctions de communication
    def query(self, text):
        """Envoyer une requête arbitraire et recevoir les 4096 premiers bits
        de la réponse."""
        if self.isconnected:
            self.lock.acquire()
            try:
                self.send(text + "\n")
                reply = self.recv(4096).strip("\n")
                self.lock.release()
                return reply
            except:
                self.lock.release()
                self.debug()
                self.disconnected()
                return ""
        else:
            return ""

    def listquery(self, text):
        """Envoyer une requête arbitraire dont la réponse est une liste."""
        replist = []
        if self.isconnected:
            self.lock.acquire()
            try:
                self.send(text + "\n")
                buff = self.recv(4096)
                count = int("0" + buff.split("\n", 1)[0])
                while buff.count("\n") != count + 1 and self.isconnected:
                    buff += self.recv(4096)
                self.lock.release()
                if self.isconnected:
                    replist = buff.split("\n")[1:-1]
            except:
                self.lock.release()
                self.debug()
                self.disconnected()
        return replist

    def recv(self, size):
        """Reçoit un message, via ssl si activé."""
        if self.isconnected:
            try:
                if self.isssl:
                    input = self.ssl.read(size)
                else:
                    input = self.socket.recv(size)
                if input == "":
                    self.disconnected()
            except:
                self.debug()
                self.disconnected()
                input = ""
        else:
            input = ""
        return input

    def send(self, text):
        """Envoie un message, via ssl si activé."""
        if self.isconnected:
            try:
                if self.isssl:
                    return self.ssl.write(text)
                else:
                    return self.socket.send(text)
            except:
                self.debug()
                self.disconnected()
                return 0
        else:
            return 0

    def sendlist(self, list):
        """Envoie une liste."""
        self.send(str(len(list)) + "\n")
        map(lambda x: self.send(str(x) + "\n"), list)

    def connect(self):
        """Connexion au serveur."""
        try:
            self.socket.connect((config_data_server, config_data_port))
            self.isconnected = True
            return True
        except socket.error:
            self.debug()
            self.disconnected()
            return False

    def disconnected(self):
        """Gestion d'une perte de la connexion."""
        self.isconnected = False
        self.acl = []
        self.socket = socket.socket()
        self.socket.settimeout(config_data_timeout)
        self.disconnect_handler()

    ### Commandes
    def hello(self):
        """Envoi d'un hello, retourne la dernière version disponible et
        si le serveur demande une mise à jour."""
        try:
            rep = self.query("hello " + config_version).split(' ', 1)
            return (rep[0] == "hello", rep[1])
        except:
            return (True, '')

    def die(self):
        """Envoi d'un die (arrête le serveur)."""
        if self.query("die") == "ok":
            self.socket.close()
            self.isconnected = False
            self.disconnected()
            return True
        else:
            return False

    def ssl(self):
        """Envoi d'une requête ssl (activation du ssl)."""
        if self.query("ssl") == "ssl":
            self.ssl = socket.ssl(self.socket)
            self.isssl = True

    def radiolist(self):
        """Envoi d'une requête radiolist."""
        try:
            radiolines = self.listquery("radiolist")
            return map(lambda n: (radiolines[3*n], radiolines[3*n+1],
                radiolines[3*n+2]), range(0, len(radiolines)/3))
        except:
            self.debug()
            return []

    def playlist(self, radio):
        """Envoi d'une requête playlist."""
        try:
            playlines = self.listquery("playlist " + radio)
            if len(playlines) == 1 and playlines[0] == "noauth":
                return ["noauth"]
            else:
                return map(lambda n: [playlines[7*n], playlines[7*n+1],
                    playlines[7*n+2], playlines[7*n+3], playlines[7*n+4],
                    playlines[7*n+5], playlines[7*n+6]],
                    range(0, len(playlines)/7))
        except:
            self.debug()
            return []

    def playlist_send(self, radio, playlist):
        """Envoi d'une playliste."""
        try:
            rep = self.query("playlistsave " + radio)
            if rep == "ok":
                self.sendlist(playlist)
                return self.recv(4096).strip("\n")
            else:
                return rep
        except:
            self.debug()
            return []

    def proglist(self, radio):
        """Envoi d'une requête proglist."""
        try:
            proglines = self.listquery("proglist " + radio)
            return map(lambda n: (proglines[3*n], proglines[3*n+1],
                proglines[3*n+2]), range(0, len(proglines)/3))
        except:
            self.debug()
            return []

    def liblist(self):
        """Envoi d'une requête liblist."""
        try:
            liblines = self.listquery("liblist")
            return map(lambda n: [liblines[7*n], liblines[7*n+1],
                liblines[7*n+2], liblines[7*n+3], liblines[7*n+4],
                liblines[7*n+5], liblines[7*n+6]], range(0, len(liblines)/7))
        except:
            self.debug()
            return []

    def skip(self, radio):
        """Envoi d'une requête skip."""
        return self.query("skip " + radio) == "ok"

    def login(self, user, password):
        """Authentification."""
        #if not self.isssl:
        #    self.ssl()
        rep = self.query("login " + user + ":" + password)
        if rep[:2] == "ok":
            self.acl = rep.split(' ', 1)[1].split(',')
            self.debug("Autorisations : " + ', '.join(self.acl))
            return True
        else:
            return False

    def __init__(self, disconnect_handler):
        """Initialisation du socket."""
        self.disconnect_handler = disconnect_handler
        self.lock = threading.Lock()
        self.isconnected = False
        self.acl = []
        self.socket = socket.socket()
        self.socket.settimeout(config_data_timeout)
        self.isssl = False


class IRCComm:
    """Classe d'abstraction de la communication IRC."""

    def debug(self, text=""):
        """Affiche des informations de debug."""
        debug(text, "IRCComm")

    def send(self, text):
        """Envoyer un message."""
        self.ircserv.privmsg(config_irc_chan, text)
        self.data_handler("<" + self.ircserv.get_nickname() + "> " + text + "\n")

    ### Fonctions de traitement des événements
    def process_once(self, source=None, condition=None):
        """Traitement des événements en attente."""
        try:
            self.irc.process_once()
        except:
            pass
        return True

    def process_pubmsg(self, conn, event):
        """Traitement d'un pubmsg."""
        for s in event.arguments():
            self.data_handler("<" + event.source().split('!')[0]
                + "> " + s + "\n")

    def process_privmsg(self, conn, event):
        """Traitement d'un privmsg."""
        for s in event.arguments():
            self.data_handler(">" + event.source().split('!')[0]
                + "< " + s + "\n")

    def connected(self, conn, event):
        """Fonction lancée lors de la réception d'un événement join."""
        self.data_handler("Connecté à " + config_irc_chan + "@"
            + config_irc_server + "!\nBienvenue sur IRC !\n")

    def disconnected(self, conn, event):
        """Fonction lancée lors d'une déconnexion."""
        self.data_handler("Déconnecté. Reconnexion dans "
            + str(config_irc_reconnect) + " secondes ...\n")
        self.irc.execute_delayed(config_irc_reconnect, self.reconnect)

    ### Fonctions de connexion
    def connect(self, nickname):
        """Se connecter."""
        self.nickname = nickname
        self.ircserv.connect(config_irc_server, config_irc_port, nickname)
        try:
            self.ircserv.join(config_irc_chan)
        except irclib.ServerNotConnectedError:
            self.debug()

    def reconnect(self):
        """Réinitialisation/reconnexion."""
        self.data_handler("Reconnexion ...\n")
        # irclib semble mal gérer la reconnexion ...
        self.__init__(self.data_handler)
        self.connect(self.nickname)

    def monitor_add(self, socket):
        """Ajouter les moniteurs sur les sockets."""
        self.ircsid = [gobject.io_add_watch(socket, gobject.IO_IN,
            self.process_once),
            gobject.io_add_watch(socket, gobject.IO_ERR, self.process_once)]

    def monitor_rm(self, socket):
        """Enlever les moniteurs sur les sockets."""
        map(gobject.source_remove, self.ircsid)

    ### Initialisation
    def __init__(self, data_handler):
        """Initialise la classe avec la fonction data_handler qui
        reçoit les lignes de texte à afficher."""
        self.irc = irclib.IRC(self.monitor_add, self.monitor_rm)
        self.ircserv = self.irc.server()
        self.data_handler = data_handler
        self.irc.add_global_handler("pubmsg", self.process_pubmsg)
        self.irc.add_global_handler("privmsg", self.process_privmsg)
        self.irc.add_global_handler("join", self.connected)
        self.irc.add_global_handler("disconnect", self.disconnected)


class PlaylistUI:
    """Classe de l'éditeur de playlistes; se gère toute seule."""

    def debug(self, text=""):
        """Affiche des informations de debug."""
        debug(text, "PlaylistUI")

    def clear(self, widget=None):
        """Vide la playliste."""
        self.liststore.clear()

    def str_duree(self, strduree):
        """Retourne une chaîne de caractères représentant une durée."""
        duree = int(strduree)
        return str(duree / 60) + ":" + ("0" + str(duree % 60))[-2:]

    ### Fonctions liées à l'édition de la liste
    def open_lib(self, widget=None):
        """Ouvre la bibliothèque pour ajouter des chansons."""
        self.library.show(self.liststore.append,
            "Ajouter des chansons à la playliste de " + self.radio_title, True)

    def delete(self, widget=None):
        """Supprime une chanson de la liste."""
        (tm, iter) = self.treeview.get_selection().get_selected()
        if iter != None:
            del self.liststore[iter]

    def save(self, widget=None):
        """Envoie la playliste au serveur."""
        playlines = reduce(lambda k, l: k + l,
            map(lambda row: map(lambda n: pickle.dumps(row[n])[:-5],
            range(0, len(row))), self.liststore))
        rep = self.servcomm.playlist_send(self.radio, playlines)
        if rep == "ok":
            self.gtk_message(gtk.MESSAGE_INFO, "Enregistrement effectué !")
        elif rep == "noauth":
            self.gtk_message(gtk.MESSAGE_ERROR,
                "Vous n'avez pas les droits sur cette playliste.")
        else:
            self.gtk_message(gtk.MESSAGE_ERROR,
                "Erreur lors de l'enregistrement.")

    ### Fonctions liées à l'interface
    def show(self, radio="", radio_title=""):
        """Charge le gestionnaire pour une playliste donnée."""
        self.clear()
        map(lambda x: self.liststore.append(
            map(lambda s: pickle.loads(s + "\np0\n."), x)),
            self.servcomm.playlist(radio))
        self.radio = radio
        self.radio_title = radio_title
        self.lblcur.set_text("<b>Playliste courante : %s</b>" % radio_title)
        self.lblcur.set_use_markup(True)
        self.window.show()
        self.window.present()

    def hide(self, widget=None):
        """Ferme la fenêtre."""
        self.window.hide()

    def button_with_icon(self, text, stock_icon):
        """Retourne un bouton avec l'icône de stock spécifiée."""
        btn = gtk.Button(text)
        image = gtk.Image()
        image.set_from_stock(stock_icon, gtk.ICON_SIZE_BUTTON)
        btn.set_image(image)
        return btn

    def delete_event(self, widget=None, event=None, data=None):
        """Gestion de la fermeture de la fenêtre."""
        self.hide()
        return True

    def gtk_message(self, type_message, texte):
        """Affiche un gtk.MessageDialog."""
        dialog = gtk.MessageDialog(None, gtk.DIALOG_DESTROY_WITH_PARENT,
            type_message, gtk.BUTTONS_CLOSE, texte)
        response = dialog.run()
        if response == gtk.RESPONSE_CLOSE:
            dialog.destroy()

    def __init__(self, library, servcomm, config):
        """Initialisation de la bibliothèque."""
        self.library = library
        self.servcomm = servcomm
        self.config = config

    ### Agencement général de la fenêtre
        window = gtk.Window(gtk.WINDOW_TOPLEVEL)

        window.set_border_width(10)
        window.set_default_size(700, 500)
        window.set_title("WRCUI :: Gestionnaire de playlistes")

        # Le vertical box qui permet de tout mettre
        globvbox = gtk.VBox(False, 0)

    ### Tableau d'affichage
        tvsw = gtk.ScrolledWindow()
        tvsw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        treeview = gtk.TreeView()
        tvsw.add(treeview)
        liststore = gtk.ListStore(str, str, str, str, str, str, str)
        treeview.set_model(liststore)

        # Cellules
        cell_genre = gtk.CellRendererText()
        cell_genre.set_property('text', True)
        cell_artist = gtk.CellRendererText()
        cell_artist.set_property('text', True)
        cell_album = gtk.CellRendererText()
        cell_album.set_property('text', True)
        cell_numero = gtk.CellRendererText()
        cell_numero.set_property('text', True)
        cell_titre = gtk.CellRendererText()
        cell_titre.set_property('text', True)
        cell_duree = gtk.CellRendererText()
        cell_duree.set_property('text', True)
        cell_uri = gtk.CellRendererText()
        cell_uri.set_property('text', True)

        # Colonnes
        column_genre = gtk.TreeViewColumn('Genre', cell_genre, text = 0)
        column_artist = gtk.TreeViewColumn('Artiste', cell_artist, text = 1)
        column_album = gtk.TreeViewColumn('Album', cell_album, text = 2)
        column_numero = gtk.TreeViewColumn('N°', cell_numero, text = 3)
        column_titre = gtk.TreeViewColumn('Titre', cell_titre, text = 4)
        column_duree = gtk.TreeViewColumn('Durée', cell_duree, text = 5)
        column_uri = gtk.TreeViewColumn('Nom de fichier', cell_uri, text = 6)

        treeview.append_column(column_genre)
        treeview.append_column(column_artist)
        treeview.append_column(column_album)
        treeview.append_column(column_numero)
        treeview.append_column(column_titre)
        treeview.append_column(column_duree)
        treeview.append_column(column_uri)

        # Options
        column_genre.set_sort_column_id(0)
        column_artist.set_sort_column_id(1)
        column_album.set_sort_column_id(2)
        column_numero.set_sort_column_id(3)
        column_titre.set_sort_column_id(4)
        column_duree.set_sort_column_id(5)
        column_uri.set_sort_column_id(6)

        column_genre.set_resizable(True)
        column_artist.set_resizable(True)
        column_album.set_resizable(True)
        column_numero.set_resizable(True)
        column_titre.set_resizable(True)
        column_duree.set_resizable(True)
        column_uri.set_resizable(True)

        treeview.set_reorderable(True)
        treeview.set_search_column(4)
        def treeview_search(model, column, key, iter):
            return not (key.lower() in reduce(lambda x,y: "%s %s" % (x,y),
                map(lambda n: model.get_value(iter, n),
                [0, 1, 2, 3, 4, 5, 6, 7])).lower())
        treeview.set_search_equal_func(treeview_search)

        treeview.show()
        tvsw.show()

    ### Titres
        lblcur = gtk.Label("<b>Aucune playliste sélectionnée !</b>")
        lblcur.set_use_markup(True)
        lblcur.show()

    ### Boutons
        btnhbox = gtk.HBox(False, 0)
        btnadd = self.button_with_icon("Ajouter ...",
            gtk.STOCK_ADD)
        btndel = self.button_with_icon("Enlever",
            gtk.STOCK_REMOVE)
        btnclear = self.button_with_icon("Vider la playliste",
            gtk.STOCK_CLEAR)
        btnsave = self.button_with_icon("Sauvegarder sur le serveur",
            gtk.STOCK_SAVE)
        btnclose = gtk.Button(gtk.STOCK_CLOSE)
        btnclose.set_use_stock(True)
        
        btnhbox.pack_start(btnadd, False, False, 0)
        btnhbox.pack_start(btndel, False, False, 0)
        btnhbox.pack_start(btnclear, False, False, 0)
        btnhbox.pack_start(btnsave, False, False, 0)
        btnhbox.pack_end(btnclose, False, False, 0)

        btnadd.show()
        btndel.show()
        btnclear.show()
        btnsave.show()
        btnclose.show()
        btnhbox.show()

    ### Agencement final
        globvbox.pack_start(tvsw, True, True, 0)
        globvbox.pack_start(lblcur, False, False, 0)
        globvbox.pack_end(btnhbox, False, False, 0)
        window.add(globvbox)

        globvbox.show()

    ### Connexion des signaux
        btnadd.connect("clicked", self.open_lib)
        btndel.connect("clicked", self.delete)
        btnclear.connect("clicked", self.clear)
        btnsave.connect("clicked", self.save)
        btnclose.connect("clicked", self.hide)

    ### Export des composants
        self.lblcur = lblcur
        self.window = window
        self.treeview = treeview
        self.liststore = liststore


class LibraryUI:
    """Classe de la bibliothèque; se gère toute seule."""

    def debug(self, text=""):
        """Affiche des informations de debug."""
        debug(text, "LibraryUI")

    def clear(self, widget=None):
        """Vide la bibliothèque."""
        self.liststore.clear()
        self.hashes = []

    def str_duree(self, strduree):
        """Retourne une chaîne de caractères représentant une durée."""
        duree = int(strduree)
        return str(duree / 60) + ":" + ("0" + str(duree % 60))[-2:]

    ### Fonctions d'importation
    def import_rhythmbox(self, widget=None):
        """Importe des chansons depuis Rhythmbox."""
        # Message d'attente
        waitdialog = gtk.Dialog("Veuillez patienter", self.window,
            gtk.DIALOG_MODAL | gtk.DIALOG_NO_SEPARATOR
            | gtk.DIALOG_DESTROY_WITH_PARENT)
        waitlabel = gtk.Label("Veuillez patienter pendant l'importation"
            + " depuis Rhythmbox.\nCeci peut prendre plusieurs secondes.")
        waitdialog.vbox.pack_start(waitlabel, True, True, 0)
        waitlabel.show()
        waitdialog.show()
        self.waitdialog = waitdialog

        # Permet d'afficher le message d'attente
        gobject.timeout_add(100, self.import_rhythmbox_xml)

    def import_rhythmbox_xml(self):
        """Analyse le fichier XML de Rhythmbox et ajoute les chansons
        à la bibliothèque."""
        # Fichier XML
        xmlpath = "%s/.gnome2/rhythmbox/rhythmdb.xml" % os.path.expanduser("~")
        xmlfile = open(xmlpath, 'r')

        # On découpe à la main le XML pour ne pas trop charger en mémoire
        # (espérons que le format de fichier ne change pas tout de suite)
        line = " "
        while line != '':
            line = xmlfile.readline().strip("\n")
            if line == '  <entry type="song">':
                buff = line
                while line != '</entry>':
                    line = xmlfile.readline().strip()
                    buff += line
                doc = xml.dom.minidom.parseString(buff)
                curinfo = {'genre': '', 'artist': '', 'album': '',
                    'track-number': '0', 'title': '', 'duration': '0', 'location': ''}
                map(lambda node: node.firstChild != None and curinfo.update(
                    {node.nodeName: node.firstChild.data}),
                    doc.firstChild.childNodes)
                newentry = [curinfo["genre"], curinfo["artist"],
                    curinfo["album"], curinfo["track-number"], curinfo["title"],
                    self.str_duree(curinfo["duration"]), "local",
                    urllib.unquote(str(curinfo["location"]))]
                self.liststore.append(newentry)
                self.hashes.append(reduce(lambda x,y: "%s %s" % (x,y),
                    newentry).lower())
                # Note pour curinfo["location"] : on utilise str pour le
                # transformer en objet str, sinon, c'est un objet unicode, et
                # urllib le décode en latin1 (enfin, c'est l'impression que
                # que ça donne) ... Ainsi on n'a pas de trucs bizarres.

        self.waitdialog.hide()

    def import_server(self, widget=None):
        """Lance l'importation de la bibliothèque du serveur."""
        # Message d'attente
        waitdialog = gtk.Dialog("Veuillez patienter", self.window,
            gtk.DIALOG_MODAL | gtk.DIALOG_NO_SEPARATOR
            | gtk.DIALOG_DESTROY_WITH_PARENT)
        waitlabel = gtk.Label("Veuillez patienter pendant l'importation"
            + " depuis le serveur.\nCeci peut prendre plusieurs secondes.")
        waitdialog.vbox.pack_start(waitlabel, True, True, 0)
        waitlabel.show()
        waitdialog.show()
        self.waitdialog = waitdialog

        # Permet d'afficher le message d'attente
        gobject.timeout_add(100, self.import_server_do)

    def import_server_do(self):
        """Importe la bibliothèque du serveur."""
        for x in self.servcomm.liblist():
            y = map(lambda s: pickle.loads(s + "\np0\n."), x)
            self.liststore.append(y[:6] + ["serveur", y[6]])
            self.hashes.append(reduce(lambda x,y: "%s %s" % (x,y), y).lower())
        self.waitdialog.hide()

    ### Fonctions de recherche
    def search(self, widget=None):
        """Filtre la bibliothèque pour les recherches."""
        inlst = self.seaentry.get_text().lower().split()
        searchlst = []
        for txt in inlst:
            if len(txt) > 1: searchlst.append(txt)
        if len(searchlst) == 0:
            self.treeview.set_model(self.liststore)
        else:
            curstore = gtk.ListStore(str, str, str, str, str, str, str, str)
            for i in range(0, len(self.hashes)):
                if reduce(lambda x,y: x and y, map(lambda x:
                        x in self.hashes[i], searchlst)):
                    curstore.append(self.liststore[i])
            self.treeview.set_model(curstore)

    ### Fonctions d'envoi
    def send_lib(self, widget=None):
        """Envoyer la chanson sélectionnée dans la bibliothèque."""
        selection = self.treeview.get_selection()
        if selection.count_selected_rows() > 0:
            curlist = self.treeview.get_model()
            def sender(tm, path, iter):
                pos = curlist.get_value(iter, 6)
                if pos == 'serveur':
                    if self.sendline:
                        self.action(map(lambda n:
                            curlist.get_value(iter, n),
                            [0, 1, 2, 3, 4, 5, 7]))
                    else:
                        self.action(curlist.get_value(iter, 7))
                else:
                    self.debug("Non implémenté : envoi depuis " + pos)
            selection.selected_foreach(sender)
        else:
            self.gtk_message(gtk.MESSAGE_ERROR,
                "Aucune chanson sélectionnée !")

    ### Fonctions liées à l'interface
    def show(self, action=None, title="", sendline=False):
        """Affiche la bibliothèque avec une action à effectuer pour
        les objets sélectionnés."""
        if action != None:
            self.action = action
            self.actok.set_label(title)
            self.acthbox.show()
        else:
            self.acthbox.hide()
        self.sendline = sendline
        self.window.show()
        self.window.present()

        # On importe automatiquement depuis le serveur (pour l'instant ...)
        if self.hastoimportfromserver:
            self.import_server()
            self.hastoimportfromserver = False

    def hide(self, widget=None):
        """Cache la fenêtre."""
        self.window.hide()

    def button_with_icon(self, text, stock_icon):
        """Retourne un bouton avec l'icône de stock spécifiée."""
        btn = gtk.Button(text)
        image = gtk.Image()
        image.set_from_stock(stock_icon, gtk.ICON_SIZE_BUTTON)
        btn.set_image(image)
        return btn

    def delete_event(self, widget=None, event=None, data=None):
        """Gestion de la fermeture de la fenêtre."""
        self.hide()
        return True

    def gtk_message(self, type_message, texte):
        """Affiche un gtk.MessageDialog."""
        dialog = gtk.MessageDialog(None, gtk.DIALOG_DESTROY_WITH_PARENT,
            type_message, gtk.BUTTONS_CLOSE, texte)
        response = dialog.run()
        if response == gtk.RESPONSE_CLOSE:
            dialog.destroy()

    def __init__(self, servcomm, config):
        """Initialisation de la bibliothèque."""
    ### Tout ce qui ne concerne pas l'UI
        self.servcomm = servcomm
        self.config = config

        self.hastoimportfromserver = True
        # Contient, pour chaque entrée, une chaîne représentant "toute" l'entrée
        self.hashes = []

    ### Agencement général de la fenêtre
        window = gtk.Window(gtk.WINDOW_TOPLEVEL)

        window.set_border_width(10)
        window.set_default_size(700, 500)
        window.set_title("WRCUI :: Bibliothèque")

        # Le vertical box qui permet de tout mettre
        globvbox = gtk.VBox(False, 0)

    ### Tableau d'affichage
        tvsw = gtk.ScrolledWindow()
        tvsw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        treeview = gtk.TreeView()
        tvsw.add(treeview)
        liststore = gtk.ListStore(str, str, str, str, str, str, str, str)
        treeview.set_model(liststore)
        treeview.get_selection().set_mode(gtk.SELECTION_MULTIPLE)

        # Cellules
        cell_genre = gtk.CellRendererText()
        cell_genre.set_property('text', True)
        cell_artist = gtk.CellRendererText()
        cell_artist.set_property('text', True)
        cell_album = gtk.CellRendererText()
        cell_album.set_property('text', True)
        cell_numero = gtk.CellRendererText()
        cell_numero.set_property('text', True)
        cell_titre = gtk.CellRendererText()
        cell_titre.set_property('text', True)
        cell_duree = gtk.CellRendererText()
        cell_duree.set_property('text', True)
        cell_pos = gtk.CellRendererText()
        cell_pos.set_property('text', True)
        cell_uri = gtk.CellRendererText()
        cell_uri.set_property('text', True)

        # Colonnes
        column_genre = gtk.TreeViewColumn('Genre', cell_genre, text = 0)
        column_artist = gtk.TreeViewColumn('Artiste', cell_artist, text = 1)
        column_album = gtk.TreeViewColumn('Album', cell_album, text = 2)
        column_numero = gtk.TreeViewColumn('N°', cell_numero, text = 3)
        column_titre = gtk.TreeViewColumn('Titre', cell_titre, text = 4)
        column_duree = gtk.TreeViewColumn('Durée', cell_duree, text = 5)
        column_pos = gtk.TreeViewColumn('Emplacement', cell_pos, text = 6)
        column_uri = gtk.TreeViewColumn('Nom de fichier', cell_uri, text = 7)

        treeview.append_column(column_genre)
        treeview.append_column(column_artist)
        treeview.append_column(column_album)
        treeview.append_column(column_numero)
        treeview.append_column(column_titre)
        treeview.append_column(column_duree)
        treeview.append_column(column_pos)
        treeview.append_column(column_uri)

        # Options
        column_genre.set_sort_column_id(0)
        column_artist.set_sort_column_id(1)
        column_album.set_sort_column_id(2)
        column_numero.set_sort_column_id(3)
        column_titre.set_sort_column_id(4)
        column_duree.set_sort_column_id(5)
        column_pos.set_sort_column_id(6)
        column_uri.set_sort_column_id(7)

        column_genre.set_resizable(True)
        column_artist.set_resizable(True)
        column_album.set_resizable(True)
        column_numero.set_resizable(True)
        column_titre.set_resizable(True)
        column_duree.set_resizable(True)
        column_pos.set_resizable(True)
        column_uri.set_resizable(True)

        treeview.set_reorderable(True)
        treeview.set_search_column(4)
        def treeview_search(model, column, key, iter):
            return not (key.lower() in reduce(lambda x,y: "%s %s" % (x,y),
                map(lambda n: model.get_value(iter, n),
                [0, 1, 2, 3, 4, 5, 6, 7])).lower())
        treeview.set_search_equal_func(treeview_search)

        treeview.show()
        tvsw.show()

    ### Boutons d'action
        acthbox = gtk.HBox(False, 0)
        impfile = self.button_with_icon("Utiliser un fichier local",
            gtk.STOCK_ADD)
        actok = self.button_with_icon("", gtk.STOCK_OK)

        acthbox.pack_start(impfile, False, False, 0)
        acthbox.pack_end(actok, True, True, 0)

        #impfile.show()
        actok.show()

    ### Champ de recherche
        seahbox = gtk.HBox(False, 0)
        sealabel = gtk.Label("Rechercher : ")
        seaentry = gtk.Entry()

        seahbox.pack_start(sealabel, False, False, 0)
        seahbox.pack_start(seaentry, True, True, 0)
        sealabel.show()
        seaentry.show()
        seahbox.show()

    ### Autres boutons
        btnhbox = gtk.HBox(False, 0)
        imprhythm = self.button_with_icon("Importer depuis Rhythmbox",
            gtk.STOCK_ADD)
        impserver = self.button_with_icon("Importer depuis le serveur",
            gtk.STOCK_ADD)
        btnclear = self.button_with_icon("Vider la bibliothèque",
            gtk.STOCK_CLEAR)
        btnclose = gtk.Button(gtk.STOCK_CLOSE)
        btnclose.set_use_stock(True)
        
        btnhbox.pack_start(imprhythm, False, False, 0)
        btnhbox.pack_start(impserver, False, False, 0)
        btnhbox.pack_start(btnclear, False, False, 0)
        btnhbox.pack_end(btnclose, False, False, 0)

        #imprhythm.show()
        #impserver.show()
        #btnclear.show()
        btnclose.show()
        btnhbox.show()

    ### Agencement final
        globvbox.pack_start(tvsw, True, True, 0)
        globvbox.pack_start(acthbox, False, False, 0)
        globvbox.pack_start(seahbox, False, False, 0)
        globvbox.pack_end(btnhbox, False, False, 0)
        window.add(globvbox)

        globvbox.show()

    ### Connexion des signaux
        actok.connect("clicked", self.send_lib)
        seaentry.connect("changed", self.search)
        imprhythm.connect("clicked", self.import_rhythmbox)
        impserver.connect("clicked", self.import_server)
        btnclear.connect("clicked", self.clear)
        btnclose.connect("clicked", self.hide)

    ### Export des composants
        self.window = window
        self.treeview = treeview
        self.liststore = liststore
        self.seaentry = seaentry
        self.actok = actok
        self.acthbox = acthbox


class UI:
    """Classe de l'interface utilisateur."""

    def match_func(self, completion, key, iter, column=0):
        """Fonction de match pour l'EntryCompletion."""
        model = completion.get_model()
        text = model.get_value(iter, column)
        try:
            return text.startswith(key)
        except:
            return False

    def button_with_icon(self, text, stock_icon):
        """Retourne un bouton avec l'icône de stock spécifiée."""
        btn = gtk.Button(text)
        image = gtk.Image()
        image.set_from_stock(stock_icon, gtk.ICON_SIZE_BUTTON)
        btn.set_image(image)
        return btn

    def togglebutton_with_icon(self, text, stock_icon):
        """Retourne un ToggleButton avec l'icône de stock spécifiée."""
        btn = gtk.ToggleButton(text)
        image = gtk.Image()
        image.set_from_stock(stock_icon, gtk.ICON_SIZE_BUTTON)
        btn.set_image(image)
        return btn

    def resize_mini(self):
        """Remet à la taille du mini-player."""
        width, height = self.window.get_size()
        self.window.resize(width, 100)

    def __init__(self):
        """Initialisation de l'interface utilisateur."""

    ### Agencement général de la fenêtre
        window = gtk.Window(gtk.WINDOW_TOPLEVEL)

        window.set_border_width(10)
        window.set_default_size(700, 100)
        window.set_title("WRCUI")
    
        # Le vertical box qui permet de tout mettre
        globvbox = gtk.VBox(False, 0)
        # Le vertical box qui contient la partie "étendue" de l'interface
        extvbox = gtk.VBox(False, 0)

    ### La zone radio
        radioframe = gtk.Frame("Radio")
        radiovbox = gtk.VBox(False, 0)
        radiohbox = gtk.HBox(False, 0)

        # Le sélecteur
        radiosel = gtk.combo_box_new_text()
        radiosel.append_text('*** Attente de la connexion au serveur ... ***')
        radiosel.set_active(0)

        # Les boutons play/stop/interface étendue
        radioplay = gtk.Button("gtk-media-play")
        radioplay.set_use_stock(True)
        radiostop = gtk.Button("gtk-media-stop")
        radiostop.set_use_stock(True)
        radioext = self.button_with_icon("Etendre",
            gtk.STOCK_FULLSCREEN)

        # Les informations
        radiostatus = gtk.Label("Aucune radio en cours de lecture.")
        radiostatus.set_alignment(0, 0)
        radioinfohbox = gtk.HBox(False, 0)
        radioinfo = gtk.Label("")
        radioinfo.set_alignment(0, 0.5)
        radioinfoextend = gtk.ToggleButton()
        radioinfoextendarrow = gtk.Arrow(gtk.ARROW_RIGHT, gtk.SHADOW_OUT)
        radioinfoextend.add(radioinfoextendarrow)
        radioinfoprev = gtk.Label("")
        radioinfoprev.set_alignment(0,0)
        radioinfonext = gtk.Label("")
        radioinfonext.set_alignment(0,0)

        # L'agencement
        radiohbox.pack_start(radiosel, True, True, 0)
        radiohbox.pack_start(radioplay, False, False, 0)
        radiohbox.pack_start(radiostop, False, False, 0)
        radiohbox.pack_end(radioext, False, False, 0)

        radioinfohbox.pack_start(radioinfoextend, False, False, 0)
        radioinfohbox.pack_end(radioinfo, True, True, 0)

        radiovbox.pack_start(radiohbox, False, False, 0)
        radiovbox.pack_start(radiostatus, False, False, 0)
        radiovbox.pack_start(radioinfohbox, False, False, 0)
        radiovbox.pack_start(radioinfoprev, False, False, 0)
        radiovbox.pack_start(radioinfonext, False, False, 0)

        radioframe.add(radiovbox)
        globvbox.pack_start(radioframe, False, False, 0)

        # On affiche
        radiosel.show()
        radioplay.show()
        radiostop.show()
        radioext.show()
        radiohbox.show()

        radiostatus.show()

        radioinfoextendarrow.show()
        radioinfoextend.show()
        radioinfo.show()
        radioinfohbox.show()

        # Les boutons supplémentaires (interface étendue)
        radiolibhbox = gtk.HBox(False, 0)
        radioauto = gtk.CheckButton("Lecture automatique au démarrage")
        radioauto.set_active(True)
        radiolibadd = self.button_with_icon("Proposer une chanson",
            gtk.STOCK_JUMP_TO)
        radiolibopen = self.button_with_icon("Ouvrir la bibliothèque",
            gtk.STOCK_OPEN)

        radiolibhbox.pack_start(radioauto, False, False, 0)
        radiolibhbox.pack_start(radiolibadd, True, True, 0)
        radiolibhbox.pack_end(radiolibopen, False, False, 0)

        radiovbox.pack_end(radiolibhbox, False, False, 0)

        radiolibadd.show()
        #radiolibopen.show()

        radiovbox.show()
        radioframe.show()

    ### Les boutons pour étendre l'interface
        exthbox = gtk.HBox(False, 0)
        extchat = gtk.ToggleButton("Chat")
        extserv = gtk.ToggleButton("Animation")

        exthbox.pack_start(extchat, False, False, 0)
        exthbox.pack_end(extserv, False, False, 0)

        extvbox.pack_start(exthbox, False, False, 0)

        exthbox.show()

    ### La zone IRC
        ircexpander = gtk.Expander("Chat IRC")
        ircframe = gtk.Frame("Chat IRC")
        ircvbox = gtk.VBox(False, 0)

        # La fenêtre de dialogue IRC
        irctext = gtk.TextView()
        irctext.set_editable(False)

        ircsw = gtk.ScrolledWindow()
        ircsw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        ircsw.add(irctext)

        ircbuff = irctext.get_buffer()
        ircbuff.set_text("Indiquez un pseudo en bas afin de vous connecter à"
            + " IRC et discuter avec les animateurs !")

        # La zone pour le pseudo
        ircnickbox = gtk.HBox(False, 0)
        ircnickin = gtk.Entry()
        ircnickok = self.button_with_icon("Se connecter", gtk.STOCK_OK)
        ircnickbox.pack_start(ircnickin, True, True, 0)
        ircnickbox.pack_end(ircnickok, False, False, 0)

        # La zone d'envoi IRC
        ircinbox = gtk.HBox(False, 0)
        ircin = gtk.Entry()
        ircsend = self.button_with_icon("Envoyer", gtk.STOCK_OK)
        ircinbox.pack_start(ircin, True, True, 0)
        ircinbox.pack_end(ircsend, False, False, 0)

        ircvbox.pack_start(ircsw, True, True, 0)
        ircvbox.pack_end(ircnickbox, False, False, 0)
        ircvbox.pack_end(ircinbox, False, False, 0)
        ircframe.add(ircvbox)
        ircexpander.add(ircframe)
        extvbox.pack_start(ircexpander, True, True, 0)

        irctext.show()
        ircsw.show()
        ircnickin.show()
        ircnickok.show()
        ircnickbox.show()
        ircin.show()
        ircsend.show()
        ircvbox.show()
        ircframe.show()
        ircexpander.show()

    ### La zone Serveur
        servexpander = gtk.Expander("Animation/Serveur")
        servframe = gtk.Frame("Serveur de contrôle")
        servvbox = gtk.VBox(False, 0)

        servstatus = gtk.Label("Etat du serveur : connexion en cours ...")
        servstatus.show()

        # La zone de login
        servloginhbox = gtk.HBox(False, 0)
        servloginlabel = gtk.Label("Login : ")
        servlogin = gtk.Entry()
        servpasslabel = gtk.Label(" Mot de passe : ")
        servpass = gtk.Entry()
        servpass.set_visibility(False)
        servloginok = self.button_with_icon("S'authentifier",
            gtk.STOCK_DIALOG_AUTHENTICATION)
        servloginhbox.pack_start(servloginlabel, False, False, 0)
        servloginhbox.pack_start(servlogin, True, True, 0)
        servloginhbox.pack_start(servpasslabel, False, False, 0)
        servloginhbox.pack_start(servpass, True, True, 0)
        servloginhbox.pack_end(servloginok, False, False, 0)

        servloginlabel.show()
        servlogin.show()
        servpasslabel.show()
        servpass.show()
        servloginok.show()
        servloginhbox.show()

        # La zone d'actions
        servacthbox = gtk.HBox(False, 0)
        servautolog = gtk.CheckButton("Connexion automatique")
        servactskip = self.button_with_icon("Skip", gtk.STOCK_MEDIA_NEXT)
        serveditpl = self.button_with_icon("Editer la playliste",
            gtk.STOCK_EDIT)
        servtograw = self.togglebutton_with_icon("Commandes raw",
            gtk.STOCK_PROPERTIES)
        servactdie = self.button_with_icon("Arrêter le serveur",
            gtk.STOCK_STOP)
        servacthbox.pack_start(servautolog, False, False, 0)
        servacthbox.pack_start(servactskip, False, False, 0)
        servacthbox.pack_start(serveditpl, False, False, 0)
        servacthbox.pack_end(servactdie, False, False, 0)
        servacthbox.pack_end(servtograw, False, False, 0)

        servautolog.show()
        servactskip.show()
        serveditpl.show()
        servtograw.show()
        servactdie.show()

        # La zone de commandes raw
        servrawhbox = gtk.HBox(False, 0)
        servrawlabel = gtk.Label("Commande raw : ")
        servrawentry = gtk.Entry()
        servrawok = self.button_with_icon("Envoyer", gtk.STOCK_OK)
        servrawhbox.pack_start(servrawlabel, False, False, 0)
        servrawhbox.pack_start(servrawentry, True, True, 0)
        servrawhbox.pack_end(servrawok, False, False, 0)

        servrawlist = gtk.ListStore(str)
        servrawcompl = gtk.EntryCompletion()
        servrawcompl.set_model(servrawlist)
        servrawcompl.set_match_func(self.match_func)
        servrawcompl.set_text_column(0)
        servrawentry.set_completion(servrawcompl)

        servrawlabel.show()
        servrawentry.show()
        servrawok.show()

        servvbox.pack_start(servstatus, False, False, 0)
        servvbox.pack_start(servloginhbox, False, False, 0)
        servvbox.pack_start(servacthbox, False, False, 0)
        servvbox.pack_end(servrawhbox, False, False, 0)
        servframe.add(servvbox)
        servexpander.add(servframe)
        extvbox.pack_start(servexpander, False, False, 0)

        servvbox.show()
        servframe.show()
        servexpander.show()

    ### Le bas de la fenêtre
        # La petite note
        wrcuinote = gtk.Label("WRCUI version " + config_version
            + " -- http://webradio.crans.org")
        # La notice de nouvelle version
        servnewver = gtk.Label()

        quitbutton = gtk.Button("gtk-quit")
        quitbutton.set_use_stock(True)

        extvbox.pack_start(wrcuinote, False, False, 0)
        extvbox.pack_end(quitbutton, False, False, 0)

        wrcuinote.show()
        quitbutton.show()

        globvbox.pack_start(extvbox, False, False, 0)
        globvbox.pack_end(servnewver, False, False, 0)

        window.add(globvbox)
        globvbox.show()
        window.show()

    ### L'icône de statut
        statusicon = gtk.StatusIcon()
        statusicon.set_from_file('/usr/share/wrc/ui/wrc.png')
        statusicon.set_tooltip("WRCUI")

    ### "Export" des composants
        # La zone radio
        self.radiosel = radiosel
        self.radioplay = radioplay
        self.radiostop = radiostop
        self.radioext = radioext
        self.radiostatus = radiostatus
        self.radioinfoextend = radioinfoextend
        self.radioinfoextendarrow = radioinfoextendarrow
        self.radioinfo = radioinfo
        self.radioinfoprev = radioinfoprev
        self.radioinfonext = radioinfonext
        self.servnewver = servnewver
        self.radioauto = radioauto
        self.radiolibadd = radiolibadd
        self.radiolibopen = radiolibopen
        self.radiolibhbox = radiolibhbox
        # La zone IRC
        self.ircexpander = ircexpander
        self.ircnickin = ircnickin
        self.ircnickok = ircnickok
        self.ircnickbox = ircnickbox
        self.ircin = ircin
        self.ircsend = ircsend
        self.ircinbox = ircinbox
        self.ircsw = ircsw
        self.ircbuff = ircbuff
        # La zone serveur
        self.servexpander = servexpander
        self.servstatus = servstatus
        self.servloginhbox = servloginhbox
        self.servlogin = servlogin
        self.servpass = servpass
        self.servloginok = servloginok
        self.servacthbox = servacthbox
        self.servautolog = servautolog
        self.servactskip = servactskip
        self.serveditpl = serveditpl
        self.servtograw = servtograw
        self.servactdie = servactdie
        self.servrawentry = servrawentry
        self.servrawok = servrawok
        self.servrawlist = servrawlist
        self.servrawhbox = servrawhbox
        # Le reste
        self.wrcuinote = wrcuinote
        self.quitbutton = quitbutton
        self.extvbox = extvbox
        self.window = window
        self.statusicon = statusicon


class Player:
    """Classe du player audio."""

    def debug(self, text=""):
        """Affiche des informations de debug."""
        debug(text, "Player")

    def start(self, uri):
        """Démarrer la lecture d'un fichier."""
        self.stop()
        self.uri = uri
        self.isplaying = True
        self.userplaying = True

        # Avec subprocess, les fils sont tués en même temps que WRCUI
        player = subprocess.Popen(["mplayer", "-quiet", self.uri],
            stdin=subprocess.PIPE, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        self.playin = player.stdin
        self.playout = player.stdout

        readthread = threading.Thread(target=self.readout)
        readthread.start()

    def stop(self, source=None, condition=None):
        """Arrêter la lecture."""
        try:
            self.playin.write("q")
            self.playin.flush()
        except:
            pass
        self.userplaying = False
        self.isplaying = False
        return False

    def readout(self, source=None, condition=None):
        """Lit la sortie du lecteur."""
        fullline = " "
        while fullline != "":
            fullline = self.playout.readline()
            line = fullline.strip()

            # On évite les lignes de Cache fill qui pourrissent les logs
            if "Cache fill" not in line:
                self.debug(line + " ")

            # Informations envoyées au milieu d'un stream
            if "Demuxer info" in line:
                lineparts = line.split(' ', 5)
                name = lineparts[2]
                try:
                    value = lineparts[5]
                except:
                    value = ""
                self.newinfo(name, value)

            # Informations envoyées au début d'un stream
            elif "Artist: " in line or "Name: " in line:
                lineparts = line.split(' ', 1)
                name = lineparts[0].strip(':')
                try:
                    value = lineparts[1]
                except:
                    value = ""
                self.newinfo(name, value)

        self.isplaying = False

    def __init__(self, newinfo):
        """Initialisation ..."""
        self.newinfo = newinfo
        self.isplaying = False
        self.userplaying = False
        self.radio = ""


class Player_windows:
    """Classe du player audio, version Windows ... Ne fait que lancer le
    lecteur et dire à l'EPG de mettre à jour les infos régulièrement ..."""

    def debug(self, text=""):
        """Affiche des informations de debug."""
        debug(text, "Player_windows")

    def start(self, uri):
        """Démarrer la lecture d'un fichier."""
        self.stop()
        self.uri = uri
        self.isplaying = True
        self.userplaying = True
        self.debug("Execution de MPlayer")
        self.pid = subprocess.Popen(["mplayer.exe", self.uri]).pid
        gobject.timeout_add(2000, self.epgupdate)

    def stop(self, source=None, condition=None):
        """Arrêter la lecture."""
        if self.isplaying:
            os.system("taskkill /f /pid " + str(self.pid))
            self.isplaying = False
        self.userplaying = False
        return False

    def epgupdate(self):
        if self.isplaying:
            self.updateinfo()
            return True
        else:
            return False

    def __init__(self, updateinfo):
        """Initialisation ..."""
        self.updateinfo = updateinfo
        self.isplaying = False
        self.userplaying = False
        self.radio = ""


class EPG:
    """Classe du guide des programmes."""

    def debug(self, text=""):
        """Affiche des informations de debug."""
        debug(text, "EPG")

    def newinfo(self, name, value):
        """Récupère les informations venant du lecteur."""
        if name == "Artist":
            self.debug("Newinfo : Artist=" + value)
            curtime = time.time()
            self.curartist = (time.time(), value)
            oldtime, curname = self.curname 
            if curtime - oldtime <= 5:
                self.updateinfo(value, curname)
        elif name == "Name":
            self.debug("Newinfo : Name=" + value)
            curtime = time.time()
            self.curname = (time.time(), value)
            oldtime, curartist = self.curartist
            if curtime - oldtime <= 5:
                self.updateinfo(curartist, value)

    def updateinfo(self, artist, name):
        """Met en correspondance les informations du serveur et celles
        du lecteur."""
        # On ne stocke pas les informations, elles sont ainsi remises à jour
        # à chaque chanson (ce qui est raisonnable)
        servinfo = map(lambda (pt, pa, pn): (float(pt) + self.offset, pa, pn),
            self.servcomm.proglist(self.curradio))

        # On cherche les chansons qui ressemblent le plus
        candidates = []
        for (ptime, partist, pname) in servinfo:
            # Bon les chansons ayant <:--:> dans leur titre ... bah euh ...
            candidates.append(partist + " <:--:> " + pname)
        corresp = difflib.get_close_matches(
            artist + " <:--:> " + name, candidates)

        if corresp != []:
            try:
                [partist, pname] = corresp[0].split(" <:--:> ", 1)
                ptime, partist, pname = servinfo[map(lambda (pt, pa, pn):
                    pa == partist and pn == pname, servinfo).index(True)]
                self.debug("Correspondance : " + partist + " -- " + pname)
                self.set_info(partist, pname)

                # On ne synchronise pas sur la première chanson
                if self.offset < -1:
                    self.offset = -1
                    newoffset = 0
                else:
                    # On met à jour le décalage serveur/client
                    newoffset = time.time() - ptime
                    self.debug("Délai : " + str(self.offset) + " += "
                        + str(newoffset) + " = " + str(self.offset + newoffset))
                    self.offset += newoffset

                # On regarde les chansons précédente et suivante
                (prevt, preva, prevn) = (0, "", "")
                (nextt, nexta, nextn) = (0, "", "")
                for (ot, oa, on) in servinfo:
                    if ot < ptime and ot > prevt:
                        (prevt, preva, prevn) = (ot + newoffset, oa, on)
                    elif ot > ptime and (nextt == 0 or ot < nextt):
                        (nextt, nexta, nextn) = (ot + newoffset, oa, on)
                self.set_prev(prevt, preva, prevn)
                self.set_next(nextt, nexta, nextn, self.offset)

            except:
                self.debug("Impossible de mettre en correspondance (erreur).")
                self.set_info(artist, name)
        else:
            self.debug("Impossible de mettre en correspondance.")
            self.set_info(artist, name)

    def updateinfo_windows(self):
        """Met à jour les informations à partir des informations du serveur
        (pour l'implémentation Windows)."""
        # On ne met à jour que si on ne connaît pas la prochaine chanson
        try:
            servinfo = self.servinfocache
        except:
            servinfo = []
        curtime = time.time()
        if True not in map(lambda (pt, pa, pn): pt > curtime, servinfo):
            servinfo = map(lambda (pt, pa, pn): (float(pt) + 10, pa, pn),
                self.servcomm.proglist(self.curradio))
            self.servinfocache = servinfo

        # On cherche les chansons courante et suivante
        (curt, cura, curn) = (0, "", "")
        (nextt, nexta, nextn) = (0, "", "")
        for (ot, oa, on) in servinfo:
            if ot < curtime and ot > curt:
                (curt, cura, curn) = (ot, oa, on)
            elif ot > curtime and (nextt == 0 or ot < nextt):
                (nextt, nexta, nextn) = (ot, oa, on)

        # On cherche la chanson précédente
        (prevt, preva, prevn) = (0, "", "")
        for (ot, oa, on) in servinfo:
            if ot < curt and ot > prevt:
                (prevt, preva, prevn) = (ot, oa, on)

        # On met à jour les informations
        self.set_info(cura, curn)
        self.set_prev(prevt, preva, prevn)
        self.set_next(nextt, nexta, nextn, -1)

    def setradio(self, radio):
        """Réinitialise avec une nouvelle radio."""
        self.curradio = radio
        self.curartist = (0, "")
        self.curname = (0, "")
        self.offset = -2

    def __init__(self, servcomm, set_info, set_prev, set_next):
        """Initialisation ..."""
        self.lock = threading.Lock()
        self.servcomm = servcomm
        self.set_info = set_info
        self.set_prev = set_prev
        self.set_next = set_next
        self.curradio = ""
        self.curartist = (0, "")
        self.curname = (0, "")
        self.servinfo = (0, [])
        self.offset = -2


class WRCUI:
    """Classe principale qui gère (presque) toutes les interactions."""

    def debug(self, text=""):
        """Affiche des informations de debug."""
        debug(text, "WRCUI")

    def has_acl(self, access, radio=''):
        """Vérifie une autorisation."""
        return ((access + " " + radio) in self.servcomm.acl
            or access in self.servcomm.acl
            or 'root' in self.servcomm.acl)

    ### Fonctions liées à la configuration
    def config_autologin(self, widget=None):
        """Modification de la connexion automatique."""
        if self.ui.servautolog.get_active():
            self.config.set("Authentification", "automatique", True)
            self.config.set("Authentification", "login",
                self.ui.servlogin.get_text())
            self.config.set("Authentification", "password",
                self.ui.servpass.get_text())
        else:
            self.config.set("Authentification", "automatique", False)
            self.config.set("Authentification", "login", "")
            self.config.set("Authentification", "password", "")

    ### Fonctions liées au guide des programmes
    def epg_setinfo(self, artist, name):
        """Affiche les informations en provenance de l'EPG sur la
        chanson courante."""
        if artist == "The Savonet Team":
            # Début de stream
            self.ui.radioinfo.set_text("Informations non disponibles.")
        else:
            self.ui.radioinfo.set_text("<b>" + artist + " -- " + name + "</b>")
            self.ui.radioinfo.set_use_markup(True)
            self.ui.window.set_title("WRCUI -- " + name)
            self.ui.statusicon.set_tooltip("WRCUI -- " + name)

    def epg_setprev(self, start, artist, name):
        """Affiche les informations en provenance de l'EPG sur la
        chanson précédente."""
        if start == 0:
            self.ui.radioinfoprev.set_text("Précédente : inconnue")
        else:
            self.ui.radioinfoprev.set_text("Précédente : " + artist + " -- " + name)

    def epg_setnext(self, start, artist, name, offset):
        """Affiche les informations en provenance de l'EPG sur la
        chanson suivante."""
        self.nextartist = artist
        self.nextname = name
        if start == 0:
            self.nextok = False
            self.ui.radioinfonext.set_text("Suivante : inconnue")
        else:
            self.nextstart = start
            if offset == -1:
                self.nextsync = False
                # On ajoute 10 secondes car c'est le délai qu'on peut
                # constater en général (enfin, plutôt 7 mais bon)
                self.nextstart += 10
                remaining = int(self.nextstart - time.time())
                self.nextsec = remaining % 60
                self.nextmin = remaining / 60
                self.ui.radioinfonext.set_text("Dans ~" + str(self.nextmin)
                    + ":" + ("0" + str(self.nextsec))[-2:] + " : "
                    + artist + " -- " + name)
            else:
                self.nextsync = True
                remaining = int(self.nextstart - time.time())
                self.nextsec = remaining % 60
                self.nextmin = remaining / 60
                self.ui.radioinfonext.set_text("Dans " + str(self.nextmin)
                    + ":" + ("0" + str(self.nextsec))[-2:] + " : "
                    + artist + " -- " + name)
            self.nextok = True

    ### Fonctions liées à la bibliothèque
    def lib_add(self, widget=None):
        """Proposer une chanson depuis la bibliothèque."""
        index = self.ui.radiosel.get_active()
        if index > 0:
            radio, title, uri = self.radiolist[index - 1]
            if self.has_acl('push', radio):
                self.library.show(self.lib_push(radio, title, False),
                    "Ajouter une chanson à la file d'attente de " + title)
            else:
                self.library.show(self.lib_push(radio, title, True),
                    "Proposer une chanson pour " + title)
        else:
            self.gtk_message(gtk.MESSAGE_ERROR, "Aucune radio sélectionnée !")

    def lib_open(self, widget=None):
        """Ouvre la bibliothèque."""
        self.library.show()

    def lib_push(self, radio, title, requesting):
        """Propose une chanson à la radio."""
        def handler(uri):
            rep = self.servcomm.query("push " + radio + " "
                + pickle.dumps(uri)[:-5])
            if requesting:
                # Mode requête
                if rep == "requested":
                    self.gtk_message(gtk.MESSAGE_INFO,
                        "Chanson proposée avec succès !")
                elif rep == "pushed":
                    self.gtk_message(gtk.MESSAGE_INFO,
                        "Le serveur a ajouté la chanson à la file d'attente."
                        + "\nPassage en mode ajout.")
                    self.library.show(self.lib_push(radio, title, False),
                        "Ajouter une chanson à la file d'attente de " + title)
                else:
                    self.gtk_message(gtk.MESSAGE_ERROR,
                        "Erreur lors de la proposition.")
            else:
                # Mode push
                if rep == "pushed":
                    self.gtk_message(gtk.MESSAGE_INFO,
                        "Chanson ajoutée à la file d'attente avec succès !")
                elif rep == "requested":
                    self.gtk_message(gtk.MESSAGE_ERROR,
                        "Le serveur a refusé le push."
                        + "\nLa chanson a été proposée à la place."
                        + "\nPassage en mode requête.")
                    self.library.show(self.lib_push(radio, title, True),
                        "Proposer une chanson pour " + title)
                else:
                    self.gtk_message(gtk.MESSAGE_ERROR,
                        "Erreur lors de l'ajout.")
        return handler

    ### Fonctions liées à la radio
    def radio_play(self, widget=None):
        """Démarrer la lecture de la radio sélectionnée."""
        if widget != None:
            # L'utilisateur a cliqué sur Lecture
            index = self.ui.radiosel.get_active()
            if index > 0:
                self.player.stop()
                self.ui.radioinfo.set_text("")
                self.ui.radioinfoprev.set_text("")
                self.ui.radioinfonext.set_text("")
                self.ui.window.set_title("WRCUI")
                self.ui.statusicon.set_tooltip("WRCUI")
                radio, title, uri = self.radiolist[index - 1]
                self.player.radio = radio
                self.player.title = title
                self.epg.setradio(radio)
                self.player.start(uri)
                self.ui.radiostatus.set_text("Lecture en cours : "
                    + title + " depuis 0:00")
                self.cursec = 0
                self.curmin = 0
                self.starttime = time.time()
                self.nextok = False
                if self.radio_update_sid:
                    gobject.source_remove(self.radio_update_sid)
                self.radio_update_sid = gobject.timeout_add(1000, self.update_radiotime)
        else:
            # On redémarre
            self.player.stop()
            self.ui.radioinfo.set_text("")
            self.ui.radioinfoprev.set_text("")
            self.ui.radioinfonext.set_text("")
            self.ui.window.set_title("WRCUI")
            self.ui.statusicon.set_tooltip("WRCUI")
            self.epg.setradio(self.player.radio)
            self.player.start(self.player.uri)
            self.ui.radiostatus.set_text("Lecture en cours : "
                + self.player.title + " depuis 0:00")
            self.cursec = 0
            self.curmin = 0
            self.starttime = time.time()
            self.nextok = False
            if self.radio_update_sid:
                gobject.source_remove(self.radio_update_sid)
            self.radio_update_sid = gobject.timeout_add(1000, self.update_radiotime)

    def radio_stop(self, widget):
        """Arrêter la lecture de la radio sélectionnée."""
        self.player.stop()
        self.ui.radiostatus.set_text("Aucune radio en cours de lecture.")
        self.ui.radioinfo.set_text("")
        self.ui.radioinfoprev.set_text("")
        self.ui.radioinfonext.set_text("")
        self.ui.window.set_title("WRCUI")

    def update_radiotime(self):
        """Ajoute une seconde au compteur de temps."""
        # Gros hack, mais j'ai essayé en parsant la sortie de MPlayer
        # et ça marche pas trop trop ... (le threading marche très très
        # mal avec GTK (méchant GTK))
        if self.player.userplaying:
            self.ui.radiostatus.set_text("Lecture en cours : "
                + self.player.title + " depuis " + str(self.curmin) + ":"
                + ("0" + str(self.cursec))[-2:])

            # Compte à rebours avant la prochaine chanson
            if self.nextok:
                if self.nextsync:
                    self.ui.radioinfonext.set_text("Dans " + str(self.nextmin)
                        + ":" + ("0" + str(self.nextsec))[-2:] + " : "
                        + self.nextartist + " -- " + self.nextname)
                else:
                    self.ui.radioinfonext.set_text("Dans ~" + str(self.nextmin)
                        + ":" + ("0" + str(self.nextsec))[-2:] + " : "
                        + self.nextartist + " -- " + self.nextname)
                self.nextsec -= 1
                if self.nextsec % 5 == 4:
                    difftime = int(self.nextstart - time.time())
                    if difftime < 0:
                        self.nextok = False
                    self.nextmin = difftime / 60
                    self.nextsec = difftime % 60

            self.cursec += 1
            # On évite que la fonction soit trop lourde
            if self.cursec % 5 == 0:
                difftime = int(time.time() - self.starttime)
                self.curmin = difftime/60
                self.cursec = difftime % 60
                if self.player.isplaying:
                    return True
                else:
                    self.ui.radiostatus.set_text(
                        "Perte de la connexion. Redémarrage dans 5 secondes.")
                    gobject.timeout_add(5000, self.radio_play)
                    return False
            else:
                return True
        else:
            return False

    def radio_info_extend(self, widget=None):
        """Afficher ou non les chansons précédente et suivante."""
        if self.ui.radioinfoextend.get_active():
            self.ui.radioinfoprev.show()
            self.ui.radioinfonext.show()
            self.ui.radioinfoextendarrow.set(gtk.ARROW_DOWN, gtk.SHADOW_OUT)
            self.config.set("Radio", "InfosEtendues", True)
        else:
            self.ui.radioinfoprev.hide()
            self.ui.radioinfonext.hide()
            self.ui.radioinfoextendarrow.set(gtk.ARROW_RIGHT, gtk.SHADOW_OUT)
            self.config.set("Radio", "InfosEtendues", False)
            self.ui.resize_mini()

    def radio_extend(self, widget=None):
        """Afficher ou non l'interface étendue."""
        if self.radio_extended:
            self.ui.radiolibhbox.hide()
            self.ui.extvbox.hide()
            self.ui.wrcuinote.hide()
            image = gtk.Image()
            image.set_from_stock(gtk.STOCK_FULLSCREEN, gtk.ICON_SIZE_BUTTON)
            self.ui.radioext.set_image(image)
            self.ui.radioext.set_label("Etendre")
            self.ui.resize_mini()
            self.radio_extended = False
        else:
            self.ui.radiolibhbox.show()
            self.ui.extvbox.show()
            self.ui.wrcuinote.show()
            image = gtk.Image()
            image.set_from_stock(gtk.STOCK_LEAVE_FULLSCREEN,
                gtk.ICON_SIZE_BUTTON)
            self.ui.radioext.set_image(image)
            self.ui.radioext.set_label("Réduire")
            self.radio_extended = True
        self.config.set("Interface", "Etendue", self.radio_extended)

    def radio_auto(self, widget=None):
        """Active ou non la lecture automatique d'une radio."""
        if self.ui.radioauto.get_active():
            index = self.ui.radiosel.get_active()
            if index > 0:
                radio, title, uri = self.radiolist[index - 1]
                self.config.set("Radio", "automatique", True)
                self.config.set("Radio", "autoid", radio)
                self.gtk_message(gtk.MESSAGE_INFO,
                    "Lecture automatique activée pour " + title + ".")
            else:
                self.gtk_message(gtk.MESSAGE_ERROR,
                    "Aucune radio sélectionnée !")
                self.ui.radioauto.set_active(False)
        else:
            self.config.set("Radio", "automatique", False)
            self.config.set("Radio", "autoid", "")

    ### Fonctions liées à IRC
    def irc_send_enter(self, widget=None, event=None):
        """Gestion de l'appui sur Enter dans la fenêtre IRC."""
        # Envoi quand on appuie sur Enter
        if event.keyval == 65293: # Enter
            self.irccomm.send(self.ui.ircin.get_text())
            self.ui.ircin.set_text("")

    def irc_send_button(self, widget=None):
        """Gestion du bouton Envoyer pour IRC."""
        self.irccomm.send(self.ui.ircin.get_text())
        self.ui.ircin.set_text("")

    def irc_receive(self, text):
        """Gestion de l'arrivée de messages IRC."""
        self.ui.ircbuff.insert(self.ui.ircbuff.get_end_iter(), text)
        vadj = self.ui.ircsw.get_vadjustment()
        vadj.set_value(vadj.upper)

    def irc_connect_enter(self, widget=None, event=None):
        """Gestion de l'appui sur Enter pour le pseudo IRC."""
        # Envoi quand on appuie sur Enter
        if event.keyval == 65293: # Enter
            self.ui.ircbuff.set_text("Connexion en cours ... (" + config_irc_server + ":" + str(config_irc_port) + ")\n")
            self.irccomm.connect(self.ui.ircnickin.get_text())
            self.ui.ircnickbox.hide()
            self.ui.ircinbox.show()

    def irc_connect_button(self, widget):
        """Gestion du bouton Se connecter pour IRC."""
        self.ui.ircbuff.set_text("Connexion en cours ... (" + config_irc_server + ":" + str(config_irc_port) + ")\n")
        self.irccomm.connect(self.ui.ircnickin.get_text())
        self.ui.ircnickbox.hide()
        self.ui.ircinbox.show()

    ### Fonctions liées à la communication avec le serveur
    def servcomm_disconnected(self):
        """Evénement de déconnexion du serveur."""
        self.ui.servacthbox.hide()
        self.ui.servtograw.set_active(False)
        self.ui.servloginhbox.show()
        gobject.timeout_add(config_data_reconnect * 1000, self.servcomm_connect)
        self.servcomm_status("déconnecté. Reconnexion dans "
            + str(config_data_reconnect) + " secondes.")

    def servcomm_connect(self):
        """Lancer la connexion avec le serveur."""
        if self.servcomm.connect():
            self.servcomm_status("connecté !")

            # Login automatique
            if self.config.getboolean("Authentification", "automatique", False):
                self.ui.servlogin.set_text(self.config.get(
                    "Authentification", "login", ""))
                self.ui.servpass.set_text(self.config.get(
                    "Authentification", "password", ""))
                self.servcomm_login()

            # Vérification de la version
            (welcome, lastver) = self.servcomm.hello()
            if lastver != config_version and lastver != '':
                # Il y a une nouvelle version disponible
                if welcome:
                    self.ui.servnewver.set_text("<b>Nouvelle version disponible : "
                        + lastver + " !</b>")
                    self.ui.servnewver.show()
                else:
                    self.ui.servnewver.set_text("<b>*** Nouvelle version disponible : "
                        + lastver + " ! Mise à jour nécessaire ! ***</b>")
                    self.ui.servnewver.show()
                self.ui.servnewver.set_use_markup(True)

            # Mise à jour de la liste des radios
            self.update_radiolist()
            if self.radiolist != []:
                if self.config.getboolean("Radio", "automatique", False):
                    if not self.radioautotriggered:
                        try:
                            self.ui.radiosel.set_active(
                                map(lambda (radio, t, u): radio ==
                                self.config.get("Radio", "autoid", ""),
                                self.radiolist).index(True)+1)
                            self.radio_play("")
                        except:
                            self.ui.radioauto.set_active(False)
                else:
                    self.ui.radioauto.set_active(False)
                self.ui.radioauto.show()
                # On ne lance automatiquement la radio qu'une seule fois
                self.radioautotriggered = True

        # Pour le timeout_add
        return False

    def servcomm_status(self, text):
        """Mettre à jour le texte de statut du serveur."""
        self.ui.servstatus.set_text("Etat du serveur : " + text)

    def update_radiolist(self):
        """Met à jour la liste des radios."""
        if self.servcomm.isconnected:
            # Radio courante
            curtitle = None
            index = self.ui.radiosel.get_active()
            if index > 0:
                curtitle = self.radiolist[index-1]

            # Mise à jour
            newradiolist = self.servcomm.radiolist()
            if newradiolist != self.radiolist:
                map(lambda x: self.ui.radiosel.remove_text(1), self.radiolist)
                map(lambda (radio, title, uri): self.ui.radiosel.append_text(title),
                    newradiolist)
                self.ui.radiosel.remove_text(0)
                self.ui.radiosel.prepend_text("Sélectionnez une radio ...")
                # On remet la liste sur la radio courante au cas où
                try:
                    self.ui.radiosel.set_active(newradiolist.index(curtitle)+1)
                except ValueError:
                    self.ui.radiosel.set_active(0)
                self.radiolist = newradiolist

            return True

    def servcomm_login(self, widget=None):
        """Authentification."""
        if self.servcomm.login(self.ui.servlogin.get_text(),
                self.ui.servpass.get_text()):
            self.ui.servloginhbox.hide()
            self.servcomm_status("authentifié en tant que "
                + self.ui.servlogin.get_text() + " !")

            # On n'affiche que les boutons autorisés
            simpleacl = map(lambda ac: ac.split(' ', 1)[0], self.servcomm.acl)
            btnmap = [('skip', self.ui.servactskip),
                      ('playlist', self.ui.serveditpl),
                      ('raw', self.ui.servtograw),
                      ('die', self.ui.servactdie)]
            for (ac, btn) in btnmap:
                if 'root' in simpleacl or ac in simpleacl:
                    btn.show()
                else:
                    btn.hide()

            self.ui.servautolog.set_active(self.config.getboolean(
                "Authentification", "automatique", False))

            self.ui.servacthbox.show()

    def servcomm_die(self, widget=None):
        """Demande l'arrêt du serveur."""
        if not self.servcomm.die():
            self.gtk_message(gtk.MESSAGE_ERROR,
                "Action refusée par le serveur.")

    def servcomm_skip(self, widget=None):
        """Saute une chanson."""
        index = self.ui.radiosel.get_active()
        if index > 0:
            radio, title, uri = self.radiolist[index - 1]
            if self.servcomm.skip(radio):
                self.gtk_message(gtk.MESSAGE_INFO,
                    "Skip effectué avec succès !")
            else:
                self.gtk_message(gtk.MESSAGE_ERROR,
                    "Action refusée par le serveur.")
        else:
            self.gtk_message(gtk.MESSAGE_ERROR, "Aucune radio sélectionnée !")

    def servcomm_cmd(self, widget=None):
        """Envoie une commande raw et afficher la réponse."""
        self.gtk_message(gtk.MESSAGE_INFO, "\n".join(
            self.servcomm.listquery("cmd " + self.ui.servrawentry.get_text())))

    def edit_playlist(self, widget=None):
        """Lance l'éditeur de playlistes."""
        index = self.ui.radiosel.get_active()
        if index > 0:
            radio, title, uri = self.radiolist[index - 1]
            self.playlistui.show(radio, title)
        else:
            self.gtk_message(gtk.MESSAGE_ERROR, "Aucune radio sélectionnée !")

    ### Fonctions liées à l'UI
    def gtk_message(self, type_message, texte):
        """Affiche un gtk.MessageDialog."""
        dialog = gtk.MessageDialog(None, gtk.DIALOG_DESTROY_WITH_PARENT,
            type_message, gtk.BUTTONS_CLOSE, texte)
        response = dialog.run()
        if response == gtk.RESPONSE_CLOSE:
            dialog.destroy()

    def statusicon_activate(self, widget=None):
        """Action de l'appui sur l'icône de statut."""
        if self.ui.window.is_active():
            self.ui.window.iconify()
        else:
            self.ui.window.present()

    def toggle_raw(self, widget=None):
        """Affiche ou non la zone de commandes raw."""
        if self.ui.servtograw.get_active():
            self.ui.servrawhbox.show()
            cmdlist = self.servcomm.listquery("cmdlist")
            self.ui.servrawlist.clear()
            for cmd in cmdlist:
                self.ui.servrawlist.append([cmd.split()[0]])
        else:
            self.ui.servrawhbox.hide()
            if not self.ui.ircexpander.get_expanded():
                # On délaie sinon il le fait trop tôt ...
                gobject.timeout_add(200, self.ui.resize_mini)

    def delete_event(self, widget=None, event=None, data=None):
        """Gestion de la fermeture de la fenêtre."""
        self.player.stop()
        gtk.main_quit()
        return False

    def irc_expanding(self, widget=None, event=None, data=None):
        """Gestion des expansions de la zone IRC."""
        self.config.set("Interface", "IRCEtendu",
            self.ui.ircexpander.get_expanded())
        if not self.ui.ircexpander.get_expanded():
            # On délaie sinon il le fait trop tôt ...
            gobject.timeout_add(200, self.ui.resize_mini)

    def serv_expanding(self, widget=None, event=None, data=None):
        """Gestion des expansions de la zone Serveur."""
        self.config.set("Interface", "ServeurEtendu",
            self.ui.servexpander.get_expanded())
        if not self.ui.ircexpander.get_expanded():
            # On délaie sinon il le fait trop tôt ...
            gobject.timeout_add(200, self.ui.resize_mini)

    ### Initialisation
    def __init__(self):
        """Initialisation ..."""
        # Initialisation de la configuration
        self.config = Config(config_configfile)

        # Initialisation de l'interface utilisateur
        ui = UI()
        self.ui = ui
        ui.window.connect("delete_event", self.delete_event)
        ui.radioplay.connect("clicked", self.radio_play)
        ui.radiostop.connect("clicked", self.radio_stop)
        ui.radioext.connect("clicked", self.radio_extend)
        ui.radioinfoextend.connect("clicked", self.radio_info_extend)
        ui.radioauto.connect("toggled", self.radio_auto)
        ui.radiolibadd.connect("clicked", self.lib_add)
        ui.radiolibopen.connect("clicked", self.lib_open)
        ui.ircnickin.connect("key-press-event", self.irc_connect_enter)
        ui.ircnickok.connect("clicked", self.irc_connect_button)
        ui.ircin.connect("key-press-event", self.irc_send_enter)
        ui.ircsend.connect("clicked", self.irc_send_button)
        ui.servloginok.connect("clicked", self.servcomm_login)
        ui.servautolog.connect("toggled", self.config_autologin)
        ui.servactskip.connect("clicked", self.servcomm_skip)
        ui.serveditpl.connect("clicked", self.edit_playlist)
        ui.servtograw.connect("toggled", self.toggle_raw)
        ui.servactdie.connect("clicked", self.servcomm_die)
        ui.servrawok.connect("clicked", self.servcomm_cmd)
        ui.quitbutton.connect("clicked", self.delete_event)
        ui.statusicon.connect("activate", self.statusicon_activate)
        ui.ircexpander.connect("notify::expanded", self.irc_expanding)
        ui.servexpander.connect("notify::expanded", self.serv_expanding)

        ui.radioinfoextend.set_active(self.config.getboolean(
            "Radio", "InfosEtendues", False))
        ui.ircexpander.set_expanded(self.config.getboolean(
            "Interface", "IRCEtendu", False))
        ui.servexpander.set_expanded(self.config.getboolean(
            "Interface", "ServeurEtendu", False))
        ui.statusicon.set_visible(True)

        # Initialisation des communications
        self.servcomm = ServerComm(self.servcomm_disconnected)
        self.radiolist = []
        self.acl = []
        self.radioautotriggered = False

        self.irccomm = IRCComm(self.irc_receive)

        # Initialisation de la bibliothèque et du guide des programmes
        self.library = LibraryUI(self.servcomm, self.config)
        self.playlistui = PlaylistUI(self.library, self.servcomm, self.config)

        self.epg = EPG(self.servcomm, self.epg_setinfo, self.epg_setprev,
            self.epg_setnext)

        # Initialisation du player
        if sys.platform != "win32":
            self.player = Player(self.epg.newinfo)
        else:
            self.player = Player_windows(self.epg.updateinfo_windows)
        self.radio_update_sid = None
        self.radio_restarting = False
        self.radio_extended = False
        if self.config.getboolean("Interface", "Etendue", False):
            self.ui.radioext.clicked()

        self.servcomm_connect()

        # Mise en place des timers
        gobject.timeout_add(300000, self.update_radiolist)
        # Au cas où ...
        gobject.timeout_add(1000, self.irccomm.process_once)


def main():
    """Lancement du programme."""
    # On initialise tout ...
    WRCUI()
    
    # La super fonction super magique qui m'aurait rendu la vie plus facile si
    # je l'avais trouvée plus tôt ... Permet d'utiliser des Threads sans
    # passer par GTK ...
    if sys.platform != "win32":
        # Bien sûr, ça ne marche pas sous Windows ...
        gtk.gdk.threads_init()

    # On lance tout !
    gtk.main()
    return 0


if __name__ == "__main__":
    main()
